type op = Add | Sub | Mult | Div | Mod |  Equal | Neq | Less | Leq | Greater | Geq | Dot

type expr =
    Literal of int
  | Id of string
  | Binop of expr * op * expr
  | Assign of expr * expr
  | Call of string * expr list
  | Array of expr * expr
  | MemberAccess of expr * op * string
  | Noexpr
  | String of string



type datatypes =  Int | Arr of expr | Struct of string | Err | Void


type var_decl = {
  vname: string;
  vtype: datatypes list;
}


type stmt =
    Block of stmt list
  | Expr of expr
  | Return of expr
  | If of expr * stmt * stmt
  | For of expr * expr * expr * stmt
  | PMethod of string * string * stmt list
  | While of expr * stmt

type datatype_decl = {
	ctype: string;
	cname: string;
}




type func_decl = {	
    fname : string;
    formals: var_decl list;
    locals : var_decl list;
    body : stmt list;
    ret : datatypes list
  }




type struct_decl = {
  sname: string;
  smembers: var_decl list
}

type program = {
  sdecls : struct_decl list;
  gdecls : var_decl list;
  fdecls : func_decl list
}

let rec string_of_expr = function
    Literal(l) -> string_of_int l
  | Id(s) -> s
  | Binop(e1, o, e2) ->
      string_of_expr e1 ^ " " ^
      (match o with
	Add -> "+" | Sub -> "-" | Mult -> "*" | Div -> "/"
      | Equal -> "==" | Neq -> "!="
      | Less -> "<" | Leq -> "<=" | Greater -> ">" | Geq -> ">=") ^ " " ^
      string_of_expr e2
  | Assign(v, e) -> string_of_expr v ^ " = " ^ string_of_expr e
  | Call(f, el) ->
      f ^ "(" ^ String.concat ", " (List.map string_of_expr el) ^ ")"
  | Noexpr -> ""

let rec string_of_stmt = function
    Block(stmts) ->
      "{\n" ^ String.concat "" (List.map string_of_stmt stmts) ^ "}\n"
  | Expr(expr) -> string_of_expr expr ^ ";\n";
  | Return(expr) -> "return " ^ string_of_expr expr ^ ";\n";
  | If(e, s, Block([])) -> "if (" ^ string_of_expr e ^ ")\n" ^ string_of_stmt s
  | If(e, s1, s2) ->  "if (" ^ string_of_expr e ^ ")\n" ^
      string_of_stmt s1 ^ "else\n" ^ string_of_stmt s2
  | For(e1, e2, e3, s) ->
      "for (" ^ string_of_expr e1  ^ " ; " ^ string_of_expr e2 ^ " ; " ^
      string_of_expr e3  ^ ") " ^ string_of_stmt s

let string_of_vdecl id =  " " ^ id.vname ^ ";\n"

(*let string_of_vdecl id = id*)

let string_of_fdecl fdecl =
  (*fdecl.fname ^ "(" ^ String.concat ", "String.concat " " fdecl.formals ^ ")\n{\n"*)
  String.concat "" (List.map string_of_vdecl fdecl.locals)
  (*String.concat "" (List.map string_of_stmt fdecl.body) ^
  "}\n"*)

let string_of_program prog = 

let globals = prog.gdecls
  and functions = prog.fdecls  in

  String.concat "" (List.map string_of_vdecl globals) ^ "\n" ^
  String.concat "\n" (List.map string_of_fdecl functions)
