%{ open Ast %}

%token SEMI LPAREN RPAREN LBRACE RBRACE COMMA
%token PLUS MINUS TIMES DIVIDE ASSIGN MOD
%token EQ NEQ LT LEQ GT GEQ
%token RETURN IF ELSE FOR WHILE BOOLEAN LIST INT CHAR STRUCT ABEGIN AEND DOT COLON VOID
%token <int> LITERAL
%token <string> ID
%token EOF
%token <string> STRING

%nonassoc NOELSE
%nonassoc ELSE
%right ASSIGN
%left EQ NEQ
%left LT GT LEQ GEQ
%left PLUS MINUS
%left TIMES DIVIDE MOD
%left DOT LPAREN RPAREN ABEGIN AEND

%start program
%type <Ast.program> program

%%

program:
        /* nothing */ { {gdecls=[];sdecls=[];fdecls=[] } }
 | program sdecl { {gdecls= $1.gdecls; sdecls=$2::$1.sdecls; fdecls= $1.fdecls} }
 | program vdecl { {gdecls= $2::$1.gdecls; sdecls=$1.sdecls; fdecls=$1.fdecls} }
 | program fdecl { {gdecls= $1.gdecls; sdecls=$1.sdecls; fdecls=$2::$1.fdecls} }
 

fdecl:
   retval formals_opt RPAREN LBRACE vdecl_list stmt_list RBRACE

     { { fname = snd $1;
         formals = $2; 
         locals = List.rev $5;
         body = List.rev $6;
         ret = fst $1
         } }
retval:
        VOID ID LPAREN { [Void], $2  }

formals_opt:
    /* nothing */ { [] }
  | formal_list   { List.rev $1 }

formal_list:
    tdecl                   { [$1] }
  | formal_list COMMA tdecl {
                  (match List.hd $3.vtype with
                  Arr(s) -> (match s with 
                    Id(id) -> raise( Failure("Array declaration: "^
                      "variable not allowed in" ^
                      "funciton argument"))
                    |_ -> $3)
                  | _ -> $3) :: $1

    
    }



vdecl_list:
    /* nothing */    { [] }
  | vdecl_list vdecl { $2 :: $1 }


vdecl:
   | tdecl SEMI { match List.hd $1.vtype with
                  Arr(s) -> (match s with
                    Noexpr -> raise( Failure("Array declaration: Size not specified"))
                    |_ -> $1)
                  | _ -> $1
                }


sdecl:
        STRUCT ID LBRACE vdecl_list RBRACE SEMI
        { {
          sname = $2;
          smembers = $4;
           }
        }


tdecl:
       INT rdecl      {
                        {
                        vname = $2.vname;
                        vtype = $2.vtype @ [Int]
                        }
                       }
                      
     | STRUCT ID rdecl {
                      { vname = $3.vname;
                        vtype = $3.vtype @ [Struct($2)]
                      }
                       }

rdecl:
        ID           {
                      { vname = $1;
                        vtype = []
                      }
                     }
        | arrdecl       { $1 }

arrdecl:
        ID ABEGIN LITERAL AEND { {
          vname = $1;
          vtype = [Arr(Literal($3))]
           } }
        | ID ABEGIN AEND { {
          vname = $1;
          vtype = [Arr(Noexpr)]
           } }
 

stmt_list:
    /* nothing */  { [] }
  | stmt_list stmt { $2 :: $1 }

stmt:
    expr SEMI { Expr($1) }
  | LBRACE stmt_list RBRACE { Block(List.rev $2) }
  | IF LPAREN expr RPAREN stmt %prec NOELSE { If($3, $5, Block([])) }
  | IF LPAREN expr RPAREN stmt ELSE stmt    { If($3, $5, $7) }
  | FOR LPAREN expr_opt SEMI expr_opt SEMI expr_opt RPAREN stmt  { For($3, $5, $7, $9) }
  | WHILE LPAREN expr RPAREN stmt { While($3, $5) }
  | ID COLON LBRACE ID COLON stmt_list RBRACE SEMI {PMethod($1,$4,(List.rev $6))}


expr_opt:
    /* nothing */ { Noexpr }
  | expr          { $1 }


expr:
   LITERAL          { Literal($1) }
  | STRING           { String($1) }
  | expr PLUS   expr { Binop($1, Add,   $3) }
  | expr MINUS  expr { Binop($1, Sub,   $3) }
  | expr TIMES  expr { Binop($1, Mult,  $3) }
  | expr MOD  expr { Binop($1, Mod,  $3) }
  | expr DIVIDE expr { Binop($1, Div,   $3) }
  | expr EQ     expr { Binop($1, Equal, $3) }
  | expr NEQ    expr { Binop($1, Neq,   $3) }
  | expr LT     expr { Binop($1, Less,  $3) }
  | expr LEQ    expr { Binop($1, Leq,   $3) }
  | expr GT     expr { Binop($1, Greater,  $3) }
  | expr GEQ    expr { Binop($1, Geq,   $3) }
  | expr ASSIGN expr   { Assign($1, $3) }
  | ID LPAREN actuals_opt RPAREN { Call($1, $3) }
  | expr DOT ID {MemberAccess($1,Dot,$3)} 
  | lvalue { $1 }
  


lvalue:
        |var  {$1}
        |LPAREN expr RPAREN {$2}




var:
        ID      { Id($1) }
        | arr   { Array( fst $1, snd $1) }

arr:
        ID ABEGIN expr AEND { Id($1),$3 }



actuals_opt:
    /* nothing */ { [] }
  | actuals_list  { List.rev $1 }

actuals_list:
    expr                    { [$1] }
  | actuals_list COMMA expr { $3 :: $1 }
