open Ast

module StringMap = Map.Make(String)

type expr_t =
  | Literal_t of int * datatypes list
  | String_t of string * datatypes list
  | Negof_t of expr_t * datatypes list
  | Id_t of string * datatypes list
  | MemberAccess_t of expr_t * op * expr_t * datatypes list
  | Array_t of expr_t * expr_t * datatypes list
  | Binop_t of expr_t * op * expr_t * datatypes list
  | Assign_t of expr_t * expr_t * datatypes list
  | Call_t of string * expr_t list  * datatypes list
  | Noexpr_t of datatypes list
 
type stmt_t =
    Block_t of stmt_t list
  | Expr_t of expr_t
  | If_t of expr_t * stmt_t * stmt_t
  | For_t of expr_t * expr_t * expr_t * stmt_t
  | While_t of expr_t * stmt_t list
  | PMethod_t of string * string * stmt_t 
  | Return_t of expr_t

 

type prog_t =
  Sast of string * expr_t list * stmt_t list


type var_entry = {
  offset:int;
  typ: datatypes list
}

type func_entry = {
  param : var_entry list;
 ret_ty : datatypes list
}

type struct_entry = {
  index: int;
  memb_index: var_entry StringMap.t
}

(* Symbol table: Information about all the names in scope *)
type envt = {
  function_index : func_entry StringMap.t; (* Index for each function *)
  struct_index   : struct_entry StringMap.t;
  global_index   : var_entry StringMap.t; (* "Address" for global variables *)
  local_index    : var_entry StringMap.t; (* FP offset for args, locals *)
}

