open Ast
open Sast
open Bytecode

let rec print_list = function 
[] -> ()
| e::l -> print_int e ; print_string " " ; print_list l;;

(*let duplicates_exist l name = List.length (List.filter (fun(x) -> x = name) l) > 1*)
let duplicates_exist l  = List.exists (fun(name)-> (List.length (List.filter (fun(x) -> x = name) l) > 1)) l;;

			let print_str_of_typs typ = match typ with
					| Int -> "Int"
					| Arr(sz) -> "Arr"
					| Void -> "Void"
					| Struct(sname) -> "Struct " ^ sname
					| Err -> "Error";;
				
			let print_type ty =
				  (List.fold_left (fun s t -> s ^ (print_str_of_typs t)) "" ty);;

let print_type_list ty = 
  (List.fold_left (fun s t -> s ^ " " ^ (print_type t)) "" ty);;



(*TODO try to fix the offsets later like professor suggested *)
let rec build_local_idx map sidx offset = (function
    [] -> map
  | hd:: tl ->
      offset := 0;
      build_local_idx
    (if StringMap.mem hd.vname map then
      raise (Failure("Double declaration of " ^ hd.vname ))
     else
      StringMap.add hd.vname
        {
          offset = 0;
          typ = hd.vtype
        } map
      )
      sidx offset tl);;


(*TODO: refactor this in one method.
 We are building another one for struct to prevent declarations in struct. Yuck i know  *)
let rec build_local_idx_struct map sidx offset = (function
    [] -> map
  | hd:: tl ->
      offset := 0;
      build_local_idx_struct
    (if StringMap.mem hd.vname map then
      raise (Failure("Double declaration of " ^ hd.vname ))
    else
	 match  hd.vtype with
		| [Int] ->
		      (StringMap.add hd.vname
			{
			  offset = 0;
			  typ = hd.vtype
			} map)
		|_ -> raise (Failure("Type " ^ print_type hd.vtype ^ " can not be declared in struct " ))
 

     )
      sidx offset tl);;

	




let check prog = 
let structs = prog.sdecls
  and globals = prog.gdecls
  and functions = prog.fdecls  in
	(* Global variable check *)
	let check_global_dbl_declaration = 
		if duplicates_exist (List.map (fun g -> g.vname) globals)
		then raise (Failure ("Duplicate definition of global variable "))

	in

	(* Struct variable check *)	
	let check_struct_dbl_declaration =
		if duplicates_exist (List.map (fun s -> s.sname) structs)
		then raise (Failure ("Duplicate definition of struct type "))
	in

	(* Function variable check *)
	
	let check_function_name_declaration =
		if duplicates_exist (List.map (fun f -> f.fname) functions)
		then raise (Failure ("Duplicate definition of function name "))
	in

	(* Function/Struct/Global combo  check *)	
	let check_function_struct_global_declaration =
		if duplicates_exist ( (List.map (fun f -> f.fname) functions) @ (List.map (fun g -> g.vname) globals)
					@  (List.map (fun s -> s.sname) structs))
		then raise (Failure ("Duplicate definition of function/struct/global name "))
	in

	let global_indexes = StringMap.empty

	in

	let struct_indexes = List.fold_left
			  (fun map stct ->
			    let soffset = ref 0 in
			      let index = build_local_idx_struct
				StringMap.empty map soffset (List.rev stct.smembers) in
				(
				  StringMap.add stct.sname
				    {
				      index = 0;
				      memb_index = index
				    } map
				)
			  )
			  StringMap.empty structs

	in

(* building function indexes also buliding out the formal structure *)
	let function_indexes = List.fold_left
			  (fun map fdecl ->
			    let rec var_to_lst ind = function
			      [] -> []
			    | hd :: tl -> ( {offset =0; typ = hd.vtype} :: (var_to_lst (ind+1) tl)) in
			      StringMap.add fdecl.fname
			      {
				param = (var_to_lst 0 fdecl.formals);
				ret_ty = fdecl.ret
			      }
			      map
			  )
			  StringMap.empty functions
	in

	(* Adding print to function indexes here*)
		let function_indexes = 
	  StringMap.add "print" 
	  {
	    param = [];
	    ret_ty = [Void]
	  }
	 function_indexes 
	in
	(* Adding print to function indexes here*)
		let function_indexes = 
	  StringMap.add "exit" 
	  {
	    param = [];
	    ret_ty = [Void]
	  }
	 function_indexes 
	in
 let get_arr_underlying_type array_id table =

		let localidxtyp = (StringMap.find array_id table).typ
		in

	 	List.hd (List.filter (fun (x) -> match x with | Arr(_) -> false
									| _ -> true) localidxtyp)

     in

(* the only binary operations that work are integers to integers *)
  let rec get_binop_result_typ t1 op t2 =
         match t1, t2, op with
        | [Int],  [Int],  _ -> [Int]
        | _ , _ , _ -> [Err]

	in
 

	let check_function env fdecl=
			 let curr_offset = ref 0 in

			 let env =
			{
				env with local_index =
				(build_local_idx StringMap.empty env.struct_index curr_offset
				(fdecl.locals @ fdecl.formals))
			}

			in
			  
			let get_func_entry name = 
			    try StringMap.find name env.function_index
			    with Not_found -> raise (Failure("Function not found: " ^ name)) 
			    in

	  let get_type_from_sast_expr = function
			    | Literal_t(i, t) -> t
			    | Negof_t(e, t) -> t
			    | Id_t(s, t) -> t
			    | MemberAccess_t(e1, r, e2, t) -> t
			    | Array_t(s, e, t) -> t
			    | Binop_t(e1, o, e2, t) -> t
			    | Assign_t(e1, e2, t) -> t
			    | Noexpr_t(t) -> t
		 	    | Call_t(s,e,t) -> t
			    | String_t(s,t) -> t
			    in

		let get_fdecl_typs name =
		    let param = (get_func_entry name).param in
		      let rec conv_param2_typ_lst = function
			[] -> []
		      | hd::tl -> hd.typ :: (conv_param2_typ_lst tl) in
		      conv_param2_typ_lst param
		 in 

			  let get_typs_from_expr_t_lst param =
			      let rec conv_el2_typ_lst = function
				[] -> []
			      | hd::tl -> get_type_from_sast_expr hd :: (conv_el2_typ_lst tl) in
			      conv_el2_typ_lst param
			    in
			  
			  let get_type_varname table varname =
			    try (StringMap.find varname table).typ
			    with Not_found -> raise (Failure("semantic.ml: Type Checking Varname not found: " ^ varname))
			    in


			
		
		  let rec list_match list1 list2 = 
		    let typ_equal t1 t2 = 
		      if t1 = t2 then true else
		      match t1, t2 with
		     | Arr(_), Arr(_) -> true 
		      | _ , _  -> false in
			match list1, list2 with
			| h1::t1, h2::t2 -> typ_equal h1 h2 && list_match t1 t2
			| [_], _ -> false
			| _, [_] -> false
			| _, _ -> true
		    in
		   let is_int_or_char ty =
		    if list_match ty [Int] then true 
		      else false
		    in
 
		  let get_assign_expr_result lh t1 rh t2 =
		    let is_lh_arr t =
		      (match (List.hd t) with 
		      | Arr(_) -> raise(Failure("Assign Type Error: Left hand side cannot be an
			array variable"))
		      | _ -> false)  in 
		   if list_match t1 t2 && not(is_lh_arr t1)  then t1
		    else
		      match t1, t2 with
		       | _ , _  -> [Err]
		    in
  let assign_result_type ty1 ty2 =
    if list_match ty1 ty2 then ty1
    else [Err]
	   in
 
	  let get_struct_table2 stct =
			    (try (StringMap.find stct env.struct_index).memb_index
			     with Not_found -> raise(Failure(" struct " ^ stct ^ " is not a type")))
			in
		  let get_struct_table typ_lst = 
		    match typ_lst with 
		    | [Struct(s)] -> (get_struct_table2 s)
			   | _ -> raise (Failure
		      ("Variable is " ^ (print_type typ_lst) ^ " and not a Struct"))

	in 
			  let rec cmp_param_typ list1 list2 fname = 

			    if (fname = "print" ) then true 
			    else
				if (List.length list1 != List.length list2)
				then
				raise(Failure("Arguments have different sizes"))
				else
			
			     ( match list1, list2 with
			      | h1::t1, h2::t2 -> 
				if (list_match (assign_result_type h1 h2) [Err]) then
				false else cmp_param_typ t1 t2 fname
			      |  [_], _ -> false
			      | _, [_] -> false
			      | _, _ -> true
				 )
			    

			in let rec check_expression table  = function
				
				Id(s) -> 
					let typ = get_type_varname table s in
					   Id_t(s, typ)

				|String(s) ->  String_t(s,[Err]) 
				| Assign (s,e) -> 
				let lh = (check_expression table s) and rh = (check_expression table e) in 
				let lh_type = get_type_from_sast_expr(lh) and rh_type = get_type_from_sast_expr(rh) in
				let ty = get_assign_expr_result lh lh_type rh rh_type in
				if list_match ty [Err] then 
				 (* Assign_t(lh, rh, [Err])*)
				  raise (Failure ("Assign mismatch: 
				   Left side is " ^ (print_type lh_type) ^ " Right
				   side is " ^ (print_type rh_type) ))
				else Assign_t(lh, rh, [Int])
				|  Literal i -> Literal_t(i, [Int])

				  | Array(base,e) -> let v1 = check_expression table e in
				    let b = check_expression table base in
				    let v1_type = get_type_from_sast_expr(v1) in
				      let btyp = get_type_from_sast_expr(b) in
				      if is_int_or_char(v1_type) then 
					Array_t(b, v1, (List.tl btyp))
				      else 
					raise (Failure ("Array index is type " ^ (print_type v1_type) ^ " and not type int")) 
				  | MemberAccess(fexpr,resolve,e) ->
	   			    let v1 = check_expression table fexpr  in
				      let v1_type = get_type_from_sast_expr(v1) in

				      let tab = (get_struct_table v1_type) in 
				      let e_e = Id(e) in 	
				      let v2 = check_expression tab e_e in
				      let v2_type = get_type_from_sast_expr(v2) in

				      (match v1_type, resolve with
					| [Struct(s)], Dot -> MemberAccess_t(v1, Dot, v2, v2_type)
					| _ , _ ->
					  raise (Failure ("Struct Error: 
					   Unknown struct with type" ^ (print_type v1_type))))

				   | Binop (e1, op, e2) -> 
				    let lh = check_expression table  e1 and rh = check_expression table e2 in
				      let lh_t = get_type_from_sast_expr(lh)
				      and rh_t = get_type_from_sast_expr(rh) in
				      let ty = get_binop_result_typ lh_t op rh_t in


					if list_match ty [Err] then
					 (* Binop_t(lh, op, rh, [Err]) *)
					  raise (Failure ("Binop mismatch: 
					   Left side is " ^ (print_type lh_t) ^ " Right
					   side is " ^ (print_type rh_t) ^ ". Binary operations are only allowed if both sides are ints")) 
					else Binop_t(lh, op, rh, ty)
				 | Call (fname, actuals) ->
					if (fname = "print" && List.length actuals > 1)
					then 
						raise (Failure ("print can not have more than one argument"))
					else (	

	   			let param_exp = List.map (check_expression table) actuals
				    and rettyp = (get_func_entry fname).ret_ty in
				    let dt = get_fdecl_typs fname in
				    let pt = get_typs_from_expr_t_lst param_exp in
				    if cmp_param_typ pt dt fname then
				      Call_t(fname, param_exp, rettyp)
				    else
				      raise (Failure ("Function " ^ fname ^ " is using arguments of
				      type " ^ (print_type_list pt) ^ " but its declaration uses type " ^ 
				      (print_type_list dt)))
				) 
 	
				| Noexpr -> Noexpr_t ([Void])

 
					  
					 
			in
			let rec check_statement table = function
		 	|     Block sl ->
				let mp = (List.map (check_statement table) sl)
				in
			    (List.fold_left (fun str lst -> str @ lst) [] mp )
			| Expr e -> [Expr_t(check_expression table e)]
			| While (cmp,stmt) ->
				let cmp_t = check_expression table cmp
				in let stmt_t = check_statement table stmt
				 in [While_t(cmp_t,stmt_t)]
			
			| For (asn, cmp, inc, b) ->
			    let asn_t = check_expression table asn and cmp_t = check_expression table cmp
			in
		    let inc_t = check_expression table  inc and stm_t = check_statement table b in
			    [For_t(asn_t, cmp_t, inc_t, Block_t(stm_t))]

		  | If (p, t, f) -> 
		    let v1 = check_expression table p and v2 = check_statement table  t and v3 = check_statement table f in
		    let v1_type = get_type_from_sast_expr(v1) in
		      if is_int_or_char(v1_type) then
			  [If_t(v1, Block_t(v2), Block_t(v3))]
		      else
			raise (Failure ("If condition is type " 
		      ^ (print_type v1_type) ^ " and not type int")) 


			| PMethod(a,b,c) ->
				let typ = get_type_varname env.local_index a
				in
				let other_typ =  get_arr_underlying_type a table	
				
				in let offset_1 = (StringMap.cardinal env.local_index) + 1
				
				in

				let p_table = (StringMap.add b
						{
							offset = offset_1;
							typ = [other_typ];

						}
						StringMap.empty)
					
			
				
				in	
				let mp = List.concat(List.map (check_statement p_table) c)
				in
					
				if List.exists (fun(typ1) -> match typ1 with  | Arr(_) -> true
									      | _ -> false	
										) typ
				then	
				[PMethod_t(a,b,Block_t(mp))]
		(*				[PMethod_t(a,b,(Block_t(t_c)))]*)
	
				else	
				 raise (Failure ("A parallel method can only be called on an array where as '"
				^ a ^ "' has type " ^ (print_type typ)))
			in	

			let check_local s =
			match s.vtype with
			|[Struct(t)] ->
				let _ = get_struct_table2 t
				in
				true
			| _ -> true
			  
			in	
			let statements =(check_statement env.local_index (Block fdecl.body)) 
		
			in	
			let _ = (List.map check_local fdecl.locals)	
			
			in statements

			(*in
			in true	
			 [Sast(fdecl.fname, (conv2_expr_t fdecl.formals), stmtblock) ] *)
			
			in
			let env = { function_index = function_indexes;
		        	   global_index   = global_indexes;
			               struct_index   = struct_indexes;
				           local_index    = StringMap.empty;
           			   }



	in


   let _ =  check_global_dbl_declaration
	    in
   let _ =  check_struct_dbl_declaration
	    in   
   let _ = check_function_name_declaration
	    in
   let _ = check_function_struct_global_declaration
	    in
   let _ = List.map (check_function env) functions
  	    in true
